#ifndef PERSISTENTMATRIX_H
#define PERSISTENTMATRIX_H
//------------------------------------------------------------------------
//
//  Name:   PersistentMatrix.h
//
//  Desc:   A class defining a Matrix that is Persistent(can be saved to a File).
//
//  Author: Joao Fabro 2014 (fabro@utfpr.edu.br)
//
//------------------------------------------------------------------------
#include <iostream>
using std::cout;
using std::ifstream;
using std::ofstream;
using std::endl;
using std::string;
#include <fstream>
//#include <random>
#include <cassert>
#include <cstdlib>
//using namespace std;
#include<string.h>


#define MAXCOLUMNS 2
#define MAXLINES 50

//JAF2014: PersistentMatrix is used for the Hill Climbing part
class PersistentMatrix
{
private:
    int id;
    string name;
    int num_columns;
    int num_lines;
    float pMatrix[MAXLINES][MAXCOLUMNS];
//    int state;//just used to know which line to "read" next

public:

    PersistentMatrix(){id=0;name="";num_columns=MAXCOLUMNS;num_lines=MAXLINES;}
    PersistentMatrix(int _nc, int _nl){PersistentMatrix();num_columns=_nc;num_lines=_nl;}

    void setId(int _id){id=_id;}

    bool setValue(int _lin, int _col, float _val){
        if ((_lin>=num_lines)||(_col>=num_columns)) return(false);
        pMatrix[_lin][_col]=_val;return true;
    }
//    void setState(int _state){state=_state;}//just used to know which line to "read" next
//    int getState(){return state;}//just used to know which line to "read" next
//    void resetState(){state=0;}//just used to know which line to "read" next

    void setName(string _name){name=_name;}
    void setNumLines(int _nl){num_lines=_nl;}
    void setNumColumns(int _nc){num_columns=_nc;}
    float getValue(int _line, int _col){return pMatrix[_line][_col];}
    void showMatrix()
    {
        for (int i = 0; i < num_lines; i++) {
            for (int j = 0; j < (num_columns); j++) {
                cout << pMatrix[i][j] << "\t";
            }
            cout << endl;
        }
    }
    bool loadControlInformation(){
        string control_suffix(".control");
        string base_name(name);
        base_name.append(control_suffix);
        const char * c = base_name.c_str();

        ifstream iMatrix (c);
        if(iMatrix  && !iMatrix.eof()){
            iMatrix>>id;
            iMatrix>>num_lines;
            iMatrix>>num_columns;
        }
        cout << "LoadControlInformation for matrix " << name << " into file " << base_name <<
             " : id=" << id << " num_lines=" << num_lines << " num_columns=" << num_columns << endl;
        iMatrix.close();
        return(true);
    }

    bool saveControlInformation(){
        char space = '\t';
        string control_suffix(".control");
        string base_name(name);
        base_name.append(control_suffix);
        const char * c = base_name.c_str();
        cout << "SaveControlInformation for matrix " << name << " into file " << base_name <<
             " : id=" << id << " num_lines=" << num_lines << " num_columns=" << num_columns << endl;

        ofstream oMatrixQS(c);
        oMatrixQS<<id<<space;
        oMatrixQS<<num_lines<<space;;
        oMatrixQS<<num_columns<<space;;
        oMatrixQS<<endl;
        oMatrixQS.close();
        return(true);
    }

    bool loadMatrix(){
        const char * c = name.c_str();
        loadControlInformation();
        ifstream iMatrix (c);
        while ( iMatrix  && !iMatrix.eof()) {
              for (int i = 0; i < num_lines; i++) {
                  for (int j = 0; j < num_columns; j++) {
                      iMatrix>>pMatrix[i][j];
                  }
              }
         }

         iMatrix.close();

         return(true);
    }

    bool saveMatrix(){
        const char * c = name.c_str();
        ofstream oMatrixQS(c);
        char space = '\t';
        for (int i = 0; i < num_lines; i++) {
            for (int j = 0; j < num_columns; j++) {
                oMatrixQS<< pMatrix[i][j]<<space;
            }
            oMatrixQS<<endl;
        }
        oMatrixQS.close();
        saveControlInformation();
        return(true);
    }

};

#endif // PERSISTENTMATRIX_H
